@@ -6,31 +6,34 @@ Python.
 .. _GNTP: http://www.growlforwindows.com/gfw/help/gntp.aspx
 """
 
+from __future__ import unicode_literals
 import hashlib
 import io
 import re
 import socket
 
+from .compat import text_type
 
-__version__ = '0.2'
+
+__version__ = '0.3'
 __all__ = ['notify', 'Notifier', 'RawIcon', 'SocketCallback', 'Event']
 
 
 SUPPORTED_VERSIONS = ['1.0']
 MAX_MESSAGE_SIZE = 4096
 MAX_LINE_SIZE = 1024
-LINE_DELIMITER = '\r\n'
-SECTION_DELIMITER = SECTION_BODY_START = SECTION_BODY_END = '\r\n'
-MESSAGE_DELIMITER = '\r\n\r\n'
+LINE_DELIMITER = b'\r\n'
+SECTION_DELIMITER = SECTION_BODY_START = SECTION_BODY_END = b'\r\n'
+MESSAGE_DELIMITER = b'\r\n\r\n'
 RESPONSE_INFORMATION_LINE_RE = re.compile(
-    'GNTP/([^ ]+) (-OK|-ERROR|-CALLBACK) NONE')
+    b'GNTP/([^ ]+) (-OK|-ERROR|-CALLBACK) NONE')
 
 
 def notify(app_name, event_name, title, text=''):
-    """Register notifier and notify event at a time.
+    """Register a notifier and send a notification at a time.
 
     :param app_name: the name of the application.
-    :param event_name: the name of the event.
+    :param event_name: the name of the notification.
     :param title: the title of the notification.
     :param text: the text of the notification.  Defaults to ``''``.
     """
@@ -47,24 +50,25 @@ class Notifier(object):
     These methods can accept the optional final callback as `callback` keyword
     argument, which run after closing the connection with the GNTP server.
 
-    `event_defs` is list of ``str``, ``unicode``, double or :class:`Event`
-    instance.  It is converted to list of :class:`Event` instance as follows:
+    `event_defs` is a list of ``str``, ``unicode``, double (of ``str`` and
+    ``bool``) or :class:`Event` instance.  It is converted to a list of
+    :class:`Event` instance as follows:
     ``str`` or ``unicode`` item becomes value of the `name` attribute of
     :class:`Event` instance, whose other attributes are defaults.  Double item
     is expanded to (`name`, `enabled`) tuple, and those values are passed to
     :class:`Event` constructor.  :class:`Event` instance item is used directly.
 
     Optional keyword arguments are passed to the `gntp_client_class`
-    constructor .
+    constructor.
 
     :param name: the name of the application.
-    :param event_defs: the definitions of the events.
-    :param icon: :class:`Icon` instance or string for the application icon.
-                 Defaults to `None`.
+    :param event_defs: the definitions of the notifications.
+    :param icon: url string or an instance of subclass of :class:`BaseIcon`
+                 for the icon of the application.  Defaults to `None`.
     :param gntp_client_class: GNTP client class.  If it is `None`,
                               :class:`GNTPClient` is used.  Defaults to `None`.
 
-    .. note:: In Growl 1.3.3, `icon` of string does not work.
+    .. note:: In Growl 1.3.3, `icon` of url string does not work.
     """
 
     def __init__(self, name, event_defs, icon=None, gntp_client_class=None,
@@ -76,53 +80,50 @@ class Notifier(object):
             gntp_client_class = GNTPClient
         self.gntp_client = gntp_client_class(**kwargs)
 
-    def register(self, _message_callback=None, callback=None):
-        """Register this notifier to GNTP server.
+    def register(self, callback=None):
+        """Register this notifier to the GNTP server.
 
-        :param _message_callback: only for use in unit testing.
         :param callback: the callback run after closing the connection with
-                         the GNTP server.
+                         the GNTP server.  Defaults to `None`.
         """
         request = RegisterRequest(self.name, self.icon, self.events)
-        self.gntp_client.register(request, _message_callback, callback)
+        self.gntp_client.register(request, callback)
 
     def notify(self, name, title, text='', id_=None, sticky=False,
                priority=0, icon=None, coalescing_id=None, callback=None,
-               _message_callback=None, _callback_message_callback=None,
                gntp_callback=None, **socket_callback_options):
-        """Notify notification to GNTP server.
+        """Send a notification to the GNTP server.
 
-        :param name: the name of the event.
+        :param name: the name of the notification.
         :param title: the title of the notification.
         :param text: the text of the notification.  Defaults to `''`.
-        :param id_: the unique ID for the notification.  Defaults to `None`.
-        :param sticky: if the notification should remain displayed until
-                       dismissed by the user.  Defaults to `False`.
+        :param id_: the unique ID for the notification.  If set, this should be
+                    unique for every request.  Defaults to `None`.
+        :param sticky: if set to `True`, the notification remains displayed
+                       until dismissed by the user.  Defaults to `False`.
         :param priority: the display hint for the receiver which may be
                          ignored.  A higher number indicates a higher priority.
                          Valid values are between -2 and 2, defaults to `0`.
-        :param icon: :class:`Icon` instance or string to display with the
-                     notification.  Defaults to `None`.
-        :param coalescing_id: if present, should contain the value of the
-                              `id_` of a previously-sent notification.
+        :param icon: url string or an instance of subclass of :class:`BaseIcon`
+                     to display with the notification.  Defaults to `None`.
+        :param coalescing_id: if set, should contain the value of the `id_` of
+                              a previously-sent notification.
                               This serves as a hint to the notification system
                               that this notification should replace/update the
                               matching previous notification.  The notification
                               system may ignore this hint.  Defaults to `None`.
         :param callback: the callback run after closing the connection with
-                         the GNTP server.
-        :param _message_callback: only for use in unit testing.
-        :param _callback_message_callback: only for use in unit testing.
+                         the GNTP server.  Defaults to `None`.
         :param gntp_callback: url string for url callback or
-                              :class:`SocketCallback` for socket callback.
-                              Defaults to `None`.
-        :param socket_callback_options: the keyword arguments to used to
+                              :class:`SocketCallback` instance for socket
+                              callback.  Defaults to `None`.
+        :param socket_callback_options: the keyword arguments to be used to
                                         instantiating :class:`SocketCallback`
                                         for socket callback.  About acceptable
                                         keyword arguments,
-                                        see :func:`coerce_to_callback()`.
+                                        see :class:`SocketCallback`.
 
-        .. note:: In Growl 1.3.3, `icon` of string does not work.
+        .. note:: In Growl 1.3.3, `icon` of url string does not work.
 
         """
         notification = Notification(name, title, text, id_=id_, sticky=sticky,
@@ -131,8 +132,7 @@ class Notifier(object):
                                     gntp_callback=gntp_callback,
                                     **socket_callback_options)
         request = NotifyRequest(self.name, notification)
-        self.gntp_client.notify(request, _message_callback, callback,
-                                _callback_message_callback,
+        self.gntp_client.notify(request, callback,
                                 notification.socket_callback)
 
 
@@ -144,13 +144,14 @@ class Event(object):
     """Represent notification type.
 
     :param name: the name of the notification.
-    :param display_name: the display name of the notification, which is shown
-                         at the Applications tab of the Growl Preferences.
-                         Defaults to `None`.
+    :param display_name: the display name of the notification, which is
+                         appeared at the Applications tab of the Growl
+                         Preferences.  Defaults to `None`.
     :param enabled: indicates if the notification should be enabled by
                     default.  Defaults to `True`.
-    :param icon: :class:`Icon` instance or string, the default icon to use
-                 for notifications of this type.  Defaults to `None`.
+    :param icon: url string or an instance of subclass of :class:`BaseIcon`
+                 for the default icon to display with the notifications of this
+                 notification type.  Defaults to `None`.
 
     .. note:: In Growl 1.3.3, `icon` does not work.
     """
@@ -194,8 +195,12 @@ class BaseRequest(object):
     @property
     def message(self):
         """Return utf-8 encoded request message."""
-        return 'GNTP/1.0 %s NONE\r\n%s\r\n' % (
-            self.message_type, self.message_body)
+        return b''.join([b'GNTP/1.0 ',
+                         self.message_type.encode('utf-8'),
+                         b' NONE',
+                         LINE_DELIMITER,
+                         self.message_body,
+                         LINE_DELIMITER])
 
     @property
     def message_body(self):
@@ -218,10 +223,11 @@ class RegisterRequest(BaseRequest):
     """Represent ``REGISTER`` request.
 
     :param app_name: the name of the application.
-    :param app_icon: :class:`Icon` instance or string for the application.
-    :param events: the list of the events.
+    :param app_icon: url string or an instance of subclass of :class:`BaseIcon`
+                     for the icon of the application.
+    :param events: the list of :class:`Event` instances.
 
-    .. note:: In Growl 1.3.3, `app_icon` of string does not work.
+    .. note:: In Growl 1.3.3, `app_icon` of url string does not work.
     """
     message_type = 'REGISTER'
 
@@ -251,7 +257,12 @@ class NotifyRequest(BaseRequest):
 
 
 class Response(object):
-    """Base class for GNTP response."""
+    """Base class for GNTP response.
+
+    :param message_type: <messagetype> of the response.  `'-OK'`, `'-ERROR'` or
+                         `'-CALLBACK'`.
+    :param headers: headers of the response.
+    """
 
     def __init__(self, message_type, headers):
         self.message_type = message_type
@@ -259,13 +270,13 @@ class Response(object):
 
 
 class GNTPConnectionCallbacksMixin(object):
-    """Defines callbacks run by the GNTP connections."""
+    """Define callbacks run by the GNTP connection."""
 
     def on_register_message(self, message):
-        """Callback for ``-OK`` response of ``REGISTER`` request.
+        r"""Callback for ``-OK`` response of ``REGISTER`` request.
 
         :param message: string of response terminated by
-                        `'\\\\r\\\\n\\\\r\\\\n\\\\r\\\\n'`.
+                        `'\\r\\n\\r\\n\\r\\n'`.
         """
         try:
             response = parse_response(message, '-OK')
@@ -275,15 +286,15 @@ class GNTPConnectionCallbacksMixin(object):
             self.final_callback(response)
 
     def on_notify_message(self, message):
-        """Callback for ``-OK`` response of ``NOTIFY`` request.
+        r"""Callback for ``-OK`` response of ``NOTIFY`` request.
 
         :param message: string of response terminated by
-                        `'\\\\r\\\\n\\\\r\\\\n\\\\r\\\\n'`.
+                        `'\\r\\n\\r\\n\\r\\n'`.
         """
         try:
             response = parse_response(message, '-OK')
             if self.socket_callback is not None:
-                self.read_message(self.callback_message_callback)
+                self.read_message(self.on_callback_message)
         finally:
             if self.socket_callback is None:
                 self.close()
@@ -291,10 +302,10 @@ class GNTPConnectionCallbacksMixin(object):
             self.final_callback(response)
 
     def on_callback_message(self, message):
-        """Callback for ``-CALLBACK`` response.
+        r"""Callback for ``-CALLBACK`` response.
 
         :param message: string of response terminated by
-                        `'\\\\r\\\\n\\\\r\\\\n\\\\r\\\\n'`.
+                        `'\\r\\n\\r\\n\\r\\n'`.
         """
         try:
             callback_response = parse_response(message, '-CALLBACK')
@@ -308,12 +319,9 @@ class GNTPConnectionCallbacksMixin(object):
 class GNTPConnection(GNTPConnectionCallbacksMixin):
     """Represent the connection with the GNTP server."""
 
-    def __init__(self, address, timeout, final_callback,
-                 callback_message_callback=None, socket_callback=None):
+    def __init__(self, address, timeout, final_callback, socket_callback=None):
         self.sock = socket.create_connection(address, timeout=timeout)
         self.final_callback = final_callback
-        self.callback_message_callback = (callback_message_callback or
-                                          self.on_callback_message)
         self.socket_callback = socket_callback
 
     def write_message(self, request):
@@ -326,7 +334,7 @@ class GNTPConnection(GNTPConnectionCallbacksMixin):
         callback(message)
 
     def close(self):
-        """Close the connection."""
+        """Close the socket."""
         self.sock.close()
         self.sock = None
 
@@ -345,50 +353,35 @@ class GNTPClient(object):
         self.address = (host, port)
         self.timeout = timeout
 
-    def register(self, request, message_callback, callback):
-        """Open connection and handle register request.
+    def register(self, request, callback):
+        """Open a connection and process the register request.
 
-        :param message_callback: the callback run with the response message of
-                                 the ``REGISTER`` request.  This callback must
-                                 close the connection.
         :param callback: the final callback run after closing connection.
         """
         conn = self._connect(callback)
         conn.write_message(request)
-        conn.read_message(message_callback or conn.on_register_message)
+        conn.read_message(conn.on_register_message)
 
-    def notify(self, request, message_callback, callback,
-               callback_message_callback, socket_callback):
-        """Open connection and handle notify request.
+    def notify(self, request, callback, socket_callback):
+        """Open a connection and process the notify request.
 
-        :param message_callback: the callback run with the response message of
-                                 the ``NOTIFY`` request.  This callback must
-                                 close the connection if the ``-CALLBACK``
-                                 response will not follow..
         :param callback: the final callback run after closing connection.
-        :param callback_message_callback: the callback run with the
-                                          ``-CALLBACK`` response message.
-                                          If not `None`, must close the
-                                          connection.
         :param socket_callback: :class:`SocketCallback` instance called by
                                 `callback_message_callback`.
         """
-        conn = self._connect(
-            callback,
-            callback_message_callback=callback_message_callback,
-            socket_callback=socket_callback)
+        conn = self._connect(callback, socket_callback=socket_callback)
         conn.write_message(request)
-        conn.read_message(message_callback or conn.on_notify_message)
+        conn.read_message(conn.on_notify_message)
 
     def _connect(self, final_callback, **kwargs):
-        """Connect to GNTP server and return the connection."""
+        """Connect to the GNTP server and return the connection."""
         return self.connection_class(self.address, self.timeout,
                                      final_callback, **kwargs)
 
 
 def generate_messages(sock, size=1024):
     """Generate messages from opened socket."""
-    buf = ''
+    buf = b''
     while True:
         buf += sock.recv(size)
         if not buf:
@@ -414,14 +407,17 @@ def parse_response(message, expected_message_type=None):
             raise GNTPError('%s is not expected message type %s' % (
                     message_type, expected_message_type))
 
-        headers = dict([s.strip() for s in line.split(':', 1)]
+        headers = dict([s.strip().decode('utf-8') for s in line.split(b':', 1)]
                        for line in lines if line)
         if message_type == '-ERROR':
             raise GNTPError('%s: %s' % (headers['Error-Code'],
                                         headers['Error-Description']))
         return Response(message_type, headers)
-    except (ValueError, GNTPError) as why:
-        raise GNTPError(why.args[0], 'original message: %s' % message)
+    except ValueError as exc:
+        raise GNTPError(exc.args[0], 'original message: %r' % message)
+    except GNTPError as exc:
+        exc.args = (exc.args[0], 'original message: %r' % message)
+        raise exc
 
 
 def parse_information_line(line):
@@ -430,17 +426,18 @@ def parse_information_line(line):
     matched = RESPONSE_INFORMATION_LINE_RE.match(line)
     if matched is None:
         raise GNTPError('invalid information line: %r' % line)
-    version, message_type = matched.groups()
+    version, message_type = [s.decode('utf-8') for s in matched.groups()]
     if version not in SUPPORTED_VERSIONS:
-        raise GNTPError('version %r is not supported' % version)
+        raise GNTPError("version '%s' is not supported" % version)
     return version, message_type
 
 
 def coerce_to_events(items):
-    """Coerce list of event definitions to list of :class:`Event` instances."""
+    """Coerce the list of the event definitions to the list of :class:`Event`
+    instances."""
     results = []
     for item in items:
-        if isinstance(item, basestring):
+        if isinstance(item, (bytes, text_type)):
             results.append(Event(item, enabled=True))
         elif isinstance(item, tuple):
             name, enabled = item
@@ -466,9 +463,9 @@ class BaseIcon(object):
 
 
 class RawIcon(BaseIcon):
-    """Class for icon.
+    """Class for icon of <uniqueid> data type.
 
-    :param data: binary data of image.
+    :param data: bytes of icon image.
     """
 
     def __init__(self, data):
@@ -492,6 +489,7 @@ class RawIcon(BaseIcon):
 
 
 class URLIcon(BaseIcon):
+    """Class for icon of <url> data type."""
 
     def __init__(self, url):
         BaseIcon.__init__(self, url)
@@ -506,19 +504,24 @@ class URLIcon(BaseIcon):
 def coerce_to_icon(obj):
     if obj is None or isinstance(obj, BaseIcon):
         return obj
-    if isinstance(obj, basestring):
+    if isinstance(obj, (bytes, text_type)):
         return URLIcon(obj)
 
 
 class SocketCallback(object):
     """Base class for socket callback.
 
+    Each of the callbacks takes one positional argument, which is
+    :class:`Response` instance.
+
     :param context: value of ``Notification-Callback-Context``.
     :param context-type: value of ``Notification-Callback-Context-Type``.
                          Defaults to ``'text/plain; charset=utf-8'``.
-    :param on_click: the function run at ``CLICKED`` callback result.
-    :param on_close: the function run at ``CLOSED`` callback result.
-    :param on_timeout: the function run at ``CLOSED`` callback result.
+    :param on_click: the callback run at ``CLICKED`` callback result.
+    :param on_close: the callback run at ``CLOSED`` callback result.
+    :param on_timeout: the callback run at ``TIMEDOUT`` callback result.
+
+    .. note:: TIMEDOUT callback does not occur in my Growl 1.3.3.
     """
 
     def __init__(self, context='', context_type='text/plain; charset=utf-8',
@@ -549,8 +552,6 @@ class SocketCallback(object):
         ``Notification-Callback-Result`` value.
 
         :param response: :class:`Response` instance.
-
-        .. note:: TIMEDOUT callback does not occur in my Growl 1.3.3.
         """
         callback_result = response.headers['Notification-Callback-Result']
         delegate_map = {
@@ -565,6 +566,7 @@ class SocketCallback(object):
 
 
 class URLCallback(object):
+    """Class for url callback."""
 
     def __init__(self, url):
         self.url = url
@@ -578,7 +580,7 @@ def coerce_to_callback(gntp_callback=None, **socket_callback_options):
     :class:`SocketCallback` instance for socket callback.
 
     If `gntp_callback` is not `None`, `socket_callback_options` must be empty.
-    If `gntp_callback` is a instance of ``basestring``, then a instance of
+    Moreover, if `gntp_callback` is string, then a instance of
     :class:`URLCallback` is returned.  Otherwise, `gntp_callback` is returned
     directly.
 
@@ -591,7 +593,7 @@ def coerce_to_callback(gntp_callback=None, **socket_callback_options):
         if socket_callback_options:
             raise GNTPError('If gntp_callback is not None,'
                             ' socket_callback_options must be empty')
-        if isinstance(gntp_callback, basestring):
+        if isinstance(gntp_callback, (bytes, text_type)):
             return URLCallback(gntp_callback)
         else:
             return gntp_callback
@@ -605,41 +607,42 @@ class HeaderWriter(object):
         self.writer = writer
 
     def write_register_request(self, request):
-        self.write('Application-Name', request.app_name)
-        self.write_icon(request.app_icon, 'Application-Icon')
-        self.write('Notifications-Count', len(request.events))
+        self.write(b'Application-Name', request.app_name)
+        self.write_icon(request.app_icon, b'Application-Icon')
+        self.write(b'Notifications-Count', len(request.events))
         for event in request.events:
             self.writer.write(LINE_DELIMITER)
             self.write_event(event)
 
     def write_notify_request(self, request):
-        self.write('Application-Name', request.app_name)
+        self.write(b'Application-Name', request.app_name)
         self.write_notification(request.notification)
 
     def write_event(self, event):
-        self.write('Notification-Name', event.name)
-        self.write('Notification-Display-Name', event.display_name)
-        self.write('Notification-Enabled', event.enabled)
-        self.write_icon(event.icon, 'Notification-Icon')
+        self.write(b'Notification-Name', event.name)
+        self.write(b'Notification-Display-Name', event.display_name)
+        self.write(b'Notification-Enabled', event.enabled)
+        self.write_icon(event.icon, b'Notification-Icon')
 
     def write_notification(self, notification):
-        self.write('Notification-Name', notification.name)
-        self.write('Notification-ID', notification.id_)
-        self.write('Notification-Title', notification.title)
-        self.write('Notification-Text', notification.text)
-        self.write('Notification-Sticky', notification.sticky)
-        self.write('Notification-Priority', notification.priority)
-        self.write_icon(notification.icon, 'Notification-Icon')
-        self.write('Notification-Coalescing-ID', notification.coalescing_id)
+        self.write(b'Notification-Name', notification.name)
+        self.write(b'Notification-ID', notification.id_)
+        self.write(b'Notification-Title', notification.title)
+        self.write(b'Notification-Text', notification.text)
+        self.write(b'Notification-Sticky', notification.sticky)
+        self.write(b'Notification-Priority', notification.priority)
+        self.write_icon(notification.icon, b'Notification-Icon')
+        self.write(b'Notification-Coalescing-ID', notification.coalescing_id)
         if notification.callback is not None:
             notification.callback.write_into(self)
 
     def write_socket_callback(self, callback):
-        self.write('Notification-Callback-Context', callback.context)
-        self.write('Notification-Callback-Context-Type', callback.context_type)
+        self.write(b'Notification-Callback-Context', callback.context)
+        self.write(b'Notification-Callback-Context-Type',
+                   callback.context_type)
 
     def write_url_callback(self, callback):
-        self.write('Notification-Callback-Target', callback.url)
+        self.write(b'Notification-Callback-Target', callback.url)
 
     def write_icon(self, icon, name):
         if icon is not None:
@@ -653,12 +656,10 @@ class HeaderWriter(object):
         Otherwise it is converted with ``str`` function.
         """
         if value is not None:
-            if isinstance(value, unicode):
-                value = value.encode('utf-8')
-            elif not isinstance(value, str):
-                value = str(value)
+            if not isinstance(value, bytes):
+                value = text_type(value).encode('utf-8')
             self.writer.write(name)
-            self.writer.write(': ')
+            self.writer.write(b': ')
             self.writer.write(value)
             self.writer.write(LINE_DELIMITER)
 
@@ -684,8 +685,8 @@ class SectionWriter(object):
         self.write_icon(notification.icon)
 
     def write_raw_icon(self, icon):
-        headers = [('Identifier', icon.unique_id()),
-                   ('Length', len(icon.data))]
+        headers = [(b'Identifier', icon.unique_id()),
+                   (b'Length', len(icon.data))]
         self.write(headers, icon.data)
 
     def write_icon(self, icon):
